/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "carlfw.h"

#include "compiler.h"

static void mini_help(void)
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "\tminiboot FW-FILE MB-FILE\n");

	fprintf(stderr, "\nDescription:\n");
	fprintf(stderr, "\tFirmware concatenation utility.\n");

	fprintf(stderr, "\nParameteres:\n");
	fprintf(stderr, "\t 'FW-FILE'	= destination for the package.\n");
	fprintf(stderr, "\t 'MB-FILE'	= extra firmware image.\n");
}

int main(int argc, char *args[])
{
	struct stat file_stat;
	struct carlfw *fw = NULL;
	struct carl9170fw_otus_desc_v1 *otus_desc = NULL;
	struct carl9170fw_usb_desc_v1 *usb_desc = NULL;
	FILE *m = NULL;
	char *buf = NULL;
	size_t extra;
	int err;

	if (argc != 3) {
		err = -EINVAL;
		goto fail;
	}

	m = fopen(args[2], "r");
	if (m == NULL) {
		fprintf(stderr, "Failed to open file %s (%d).\n",
			args[2], errno);
		err = -errno;
		goto fail;
	}

	err = fstat(fileno(m), &file_stat);
	if (err) {
		fprintf(stderr, "Failed to query file infos from "
				"\"%s\" (%d).\n", args[2], errno);
		err = -errno;
		goto fail;
	}
	extra = file_stat.st_size;

	fw = carlfw_load(args[1]);
	if (IS_ERR_OR_NULL(fw)) {
		err = PTR_ERR(fw);
		goto fail;
	}

	otus_desc = carlfw_find_desc(fw, (uint8_t *) OTUS_MAGIC,
				     sizeof(*otus_desc),
				     CARL9170FW_OTUS_DESC_CUR_VER);
	if (!otus_desc) {
		fprintf(stderr, "No OTUS descriptor found in \"%s\"\n",
			args[1]);
		goto fail;
	}

	usb_desc = carlfw_find_desc(fw, (uint8_t *) USB_MAGIC,
				    sizeof(*usb_desc),
				    CARL9170FW_USB_DESC_CUR_VER);
	if (!usb_desc) {
		fprintf(stderr, "No USB descriptor found in \"%s\"\n",
			args[1]);
		goto fail;
	}

	if (carl9170fw_supports(otus_desc->fw_feature_set, AR9170FW_LEGACY)) {
		fprintf(stderr, "Firmware \"%s\" is supposed to be compatible "
			"with legacy drivers.\n", args[1]);
		goto fail;
	}

	if (carl9170fw_supports(usb_desc->usb_feature_set, CARL9170FW_USB_MINIBOOT)) {
		fprintf(stderr, "Firmware \"%s\" has already a "
				"miniboot image.\n", args[1]);
		goto fail;
	}

	usb_desc->usb_feature_set |= cpu_to_le32(BIT(CARL9170FW_USB_MINIBOOT));
	usb_desc->miniboot_size = cpu_to_le16(extra);

	buf = carlfw_mod_headroom(fw, extra);
	if (IS_ERR_OR_NULL(buf)) {
		fprintf(stderr, "Failed to make some headroom \"%s\" (%d).\n",
			args[2], err);
		goto fail;
	}

	err = fread(buf, extra, 1, m);
	if (err != 1) {
		fprintf(stderr, "Failed to read from miniboot \"%s\" (%d).\n",
			args[2], errno);
		goto fail;
	}

	carlfw_store(fw);
	fclose(m);

	fprintf(stdout, "Miniboot applied.\n");
	return EXIT_SUCCESS;

fail:
	if (m)
		fclose(m);

	switch (err) {
	case 0:
		break;

	case -EINVAL:
		mini_help();
		break;

	default:
		fprintf(stderr, "%s\n", strerror(-err));
		break;
	}

	carlfw_release(fw);
	return EXIT_FAILURE;
}
